package com.hortonworks.spark.sql.hive.llap.writers;

import com.hortonworks.spark.sql.hive.llap.util.JobUtil;
import com.hortonworks.spark.sql.hive.llap.util.SparkToHiveRecordMapper;
import com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataWriterHelper.FileFormat;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.TaskAttemptID;
import org.apache.hadoop.mapreduce.task.TaskAttemptContextImpl;
import org.apache.spark.sql.catalyst.InternalRow;
import org.apache.spark.sql.execution.datasources.BasicWriteJobStatsTracker;
import org.apache.spark.sql.execution.datasources.OutputWriter;
import org.apache.spark.sql.execution.datasources.WriteTaskStats;
import org.apache.spark.sql.execution.datasources.WriteTaskStatsTracker;
import org.apache.spark.sql.sources.v2.writer.DataWriter;
import org.apache.spark.sql.sources.v2.writer.WriterCommitMessage;
import org.apache.spark.sql.types.StructType;
import org.apache.spark.sql.types.TimestampType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Map;
import java.util.ArrayList;
import java.util.List;
import java.util.TimeZone;

public class HiveWarehouseDataWriter implements DataWriter<InternalRow> {
  private static Logger LOG = LoggerFactory.getLogger(HiveWarehouseDataWriter.class);

  private String jobId;
  private Configuration conf;
  private int partitionId;
  private long taskId;
  private long epochId;
  private FileSystem fs;
  private Path filePath;
  protected OutputWriter out;
  private SparkToHiveRecordMapper sparkToHiveRecordMapper;
  private WriteTaskStatsTracker basicWriteStats;
  private BasicWriteJobStatsTracker basicWriteJobStatsTracker;
  private HiveWarehouseDataWriterHelper HiveWarehouseDataWriterHelper;
  private final TimeZone jvmDefaultTimezone = TimeZone.getDefault();
  private final TimeZone sparkSessionTimezone;
  private final boolean needTimezoneConversion;
  private final List<Integer> timestampFieldIndices;


  public HiveWarehouseDataWriter(Configuration conf, String jobId, StructType schema, FileFormat fileFormat,
                                 Map<String, String> options, int partitionId, long taskId, long epochId,
                                 FileSystem fs, Path filePath, SparkToHiveRecordMapper sparkToHiveRecordMapper,
                                 BasicWriteJobStatsTracker basicWriteJobStatsTracker) {
    this.conf = conf;
    this.jobId = jobId;
    this.partitionId = partitionId;
    this.taskId = taskId;
    this.epochId = epochId;
    this.fs = fs;
    this.filePath = filePath;
    this.sparkToHiveRecordMapper = sparkToHiveRecordMapper;
    this.HiveWarehouseDataWriterHelper = new HiveWarehouseDataWriterHelper(conf, schema, fileFormat, options, filePath, sparkToHiveRecordMapper);

    this.conf = HiveWarehouseDataWriterHelper.getConf();
    TaskAttemptContext tac = new TaskAttemptContextImpl(this.conf, new TaskAttemptID());
    this.out = HiveWarehouseDataWriterHelper.getOutputWriter(tac);

    this.basicWriteJobStatsTracker = basicWriteJobStatsTracker;
    this.basicWriteStats=basicWriteJobStatsTracker.newTaskInstance();
    this.basicWriteStats.newFile(String.valueOf(filePath));

    this.sparkSessionTimezone = JobUtil.getSparkSessionTimezoneAtExecutor(conf);
    this.timestampFieldIndices = new ArrayList<>();
    for (int i = 0; i < schema.fields().length; i++) {
      if (schema.fields()[i].dataType() instanceof TimestampType) {
        timestampFieldIndices.add(i);
      }
    }
    this.needTimezoneConversion = !jvmDefaultTimezone.equals(sparkSessionTimezone) && !timestampFieldIndices.isEmpty();
    logDebug("CREATED...");

  }

  @Override public void write(InternalRow record) throws IOException {
    if (needTimezoneConversion) {
      for (Integer index : timestampFieldIndices) {
        record.setLong(index, JobUtil.convertMicros(record.getLong(index), sparkSessionTimezone, jvmDefaultTimezone));
      }
    }
    out.write(sparkToHiveRecordMapper.mapToHiveColumns(record));
    basicWriteStats.newRow(record);
  }

  @Override public WriterCommitMessage commit() throws IOException {
    out.close();
    WriteTaskStats stats = basicWriteStats.getFinalStats();
    LOG.debug("Stats="+stats);
    logDebug("COMMITTING...");
    return new SimpleWriterCommitMessage(String.format("COMMIT %s_%s_%s_%s", jobId, partitionId, taskId, epochId), filePath);
  }

  @Override public void abort() throws IOException {
    LOG.info("Driver sent abort for {}_{}_{}_{}", jobId, partitionId, taskId, epochId);
    logInfo("ABORT RECEIVED FROM DRIVER...");
    try {
      out.close();
    } finally {
      fs.delete(filePath, false);
    }
  }

  private void logInfo(String msg) {
    LOG.info("HiveWarehouseDataWriter: {}, path: {}, msg:{} ", this, filePath, msg);
  }

  private void logDebug(String msg) {
    LOG.debug("HiveWarehouseDataWriter: {}, path: {}, msg:{} ", this, filePath, msg);
  }

}
