/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hortonworks.spark.sql.hive.llap.writers;

import com.hortonworks.spark.sql.hive.llap.util.SparkToHiveRecordMapper;
import org.apache.avro.Schema;
import org.apache.avro.mapred.AvroKey;
import org.apache.avro.mapred.AvroOutputFormat;
import org.apache.avro.mapreduce.AvroJob;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.parquet.hadoop.ParquetOutputFormat;
import org.apache.spark.sql.avro.AvroOptions;
import org.apache.spark.sql.avro.AvroOutputWriter;
import org.apache.spark.sql.avro.SchemaConverters;
import org.apache.spark.sql.catalyst.csv.CSVOptions;
import org.apache.spark.sql.catalyst.util.CompressionCodecs;
import org.apache.spark.sql.execution.datasources.OutputWriter;
import org.apache.spark.sql.execution.datasources.csv.CsvOutputWriter;
import org.apache.spark.sql.execution.datasources.orc.OrcOptions;
import org.apache.spark.sql.execution.datasources.parquet.ParquetOptions;
import org.apache.spark.sql.execution.datasources.parquet.ParquetOutputWriter;
import org.apache.spark.sql.execution.datasources.parquet.ParquetWriteSupport;
import org.apache.spark.sql.hive.orc.OrcOutputWriter;
import org.apache.spark.sql.internal.SQLConf;
import org.apache.spark.sql.types.StructType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Map;
import java.util.TimeZone;

import static org.apache.avro.file.DataFileConstants.*;
import static org.apache.orc.OrcConf.COMPRESS;

public class HiveWarehouseDataWriterHelper {

    public enum FileFormat {
        ORC, PARQUET, AVRO, TEXT;

        public String toString() {
            switch (this) {
                case ORC:
                    return "orc";
                case PARQUET:
                    return "parquet";
                case AVRO:
                    return "avro";
                case TEXT:
                    return "textfile";
                default:
                    return null;
            }
        }

        public static FileFormat getFormat(String fileFormatAsString) {
            switch (fileFormatAsString.toLowerCase()) {
                case "orc":
                    return ORC;
                case "parquet":
                    return PARQUET;
                case "avro":
                    return AVRO;
                case "textfile":
                    return TEXT;
                default:
                    throw new IllegalArgumentException("Format '" + fileFormatAsString + "' is not a supported format.");
            }
        }
    }

    private static Logger LOG = LoggerFactory.getLogger(HiveWarehouseDataWriter.class);

    private Configuration conf;
    private StructType schema;
    private FileFormat fileFormat;
    private Map<String, String> options;
    private scala.collection.immutable.Map<String, String> optionsAsScalaMap;
    private String path;
    private TaskAttemptContext tac;
    private SparkToHiveRecordMapper sparkToHiveRecordMapper;
    private Schema outputAvroSchema;

    public HiveWarehouseDataWriterHelper(Configuration conf, StructType schema, FileFormat fileFormat, Map<String, String> options, Path filePath, SparkToHiveRecordMapper sparkToHiveRecordMapper) {
        this.conf = conf;
        this.schema = schema;
        this.fileFormat = fileFormat;
        this.path = filePath.toString();
        this.options = options;
        this.sparkToHiveRecordMapper = sparkToHiveRecordMapper;
        DataFrameWriterUtils dataFrameWriterUtils = new DataFrameWriterUtils();
        this.optionsAsScalaMap = dataFrameWriterUtils.MapConverter(options);
    }

    public void setConf() {
        switch (fileFormat) {
            case ORC:
                setORCConf();
                break;
            case PARQUET:
                setParquetConf();
                break;
            case AVRO:
                setAvroConf();
                break;
            case TEXT:
                setTextConf();
                break;
        }
    }

    public Configuration getConf() {
        setConf();
        return this.conf;
    }

    private void setTextConf() {
        String name = options.getOrDefault("compression", options.getOrDefault("codec", "none"));
        String codec = CompressionCodecs.getCodecClassName(name);
        CompressionCodecs.setCodecConfiguration(conf, codec);
    }

    private void setAvroConf() {
        AvroOptions parsedOptions = new AvroOptions(optionsAsScalaMap, conf);
        outputAvroSchema = SchemaConverters.toAvroType(schema, false, parsedOptions.recordName(), parsedOptions.recordNamespace());
        Job job = null;
        try {
            job = Job.getInstance(conf);
            job.setOutputKeyClass(AvroKey.class);
            conf = job.getConfiguration();
        } catch (IOException e) {
            e.printStackTrace();
        }
        conf.set("avro.schema.output.key", outputAvroSchema.toString());

        if (parsedOptions.compression() == "uncompressed") {
            conf.setBoolean("mapred.output.compress", false);
        } else {
            conf.setBoolean("mapred.output.compress", true);
            String codec = parsedOptions.compression();
            switch (codec) {
                case DEFLATE_CODEC:
                    int deflateLevel = Integer.parseInt(SQLConf.AVRO_DEFLATE_LEVEL().defaultValueString());
                    conf.setInt(AvroOutputFormat.DEFLATE_LEVEL_KEY, deflateLevel);
                    codec = DEFLATE_CODEC;
                    break;
                case SNAPPY_CODEC:
                case BZIP2_CODEC:
                case XZ_CODEC:
                    break;
                default:
                    throw new IllegalArgumentException("Invalid compression codec: " + codec);
            }
            conf.set(AvroJob.CONF_OUTPUT_CODEC, codec);
        }
    }

    private void setParquetConf() {
        Job job = null;
        try {
            job = Job.getInstance(conf);
            job.setOutputFormatClass(ParquetOutputFormat.class);
            conf = job.getConfiguration();
        } catch (IOException e) {
            e.printStackTrace();
        }
        ParquetOutputFormat.setWriteSupportClass(job, ParquetWriteSupport.class);
        ParquetWriteSupport.setSchema(schema, conf);
        conf.setIfUnset(SQLConf.PARQUET_WRITE_LEGACY_FORMAT().key(), "false");
        conf.setIfUnset(SQLConf.PARQUET_OUTPUT_COMMITTER_CLASS().key(), "org.apache.parquet.hadoop.ParquetOutputCommitter");
        conf.setIfUnset(SQLConf.PARQUET_OUTPUT_TIMESTAMP_TYPE().key(), "TIMESTAMP_MICROS");
        ParquetOptions parquetOptions = new ParquetOptions(optionsAsScalaMap, new SQLConf());
        conf.set(ParquetOutputFormat.COMPRESSION, parquetOptions.compressionCodecClassName());
    }

    private void setORCConf() {
        conf.set("orc.mapred.output.schema", sparkToHiveRecordMapper.getSchemaInHiveColumnsOrder().catalogString());
        OrcOptions orcOptions = new OrcOptions(optionsAsScalaMap, new SQLConf());
        conf.set(COMPRESS.getAttribute(), orcOptions.compressionCodec());
    }

    public OutputWriter getOutputWriter(TaskAttemptContext tac) {
        this.tac = tac;
        switch (fileFormat) {
            case ORC:
                return getOrcOutputWriter();
            case PARQUET:
                return getParquetOutputWriter();
            case AVRO:
                return getAvroOutputWriter();
            case TEXT:
                return getTextOutputWriter();
            default:
                throw new IllegalArgumentException("Output Writer for " + fileFormat + " is not supported");
        }
    }

    private OutputWriter getTextOutputWriter() {
        CSVOptions csvOptions = new CSVOptions(optionsAsScalaMap, true, TimeZone.getDefault().getID(), "");
        return new CsvOutputWriter(path, schema, tac, csvOptions);
    }

    private OutputWriter getAvroOutputWriter() {
        return new AvroOutputWriter(path, tac, schema, outputAvroSchema);
    }

    private OutputWriter getParquetOutputWriter() {
        return new ParquetOutputWriter(path, tac);
    }

    private OutputWriter getOrcOutputWriter() {
        return new OrcOutputWriter(path, schema, tac);
    }
}
